/* Terraform - (C) 1997-1999 Robert Gasch (Robert_Gasch@peoplesoft.com)
 *  - www.peoplesoft.com/peoplepages/g/robert_gasch/terraform/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include "TFDialogCraters.h"
#include "GlobalSanityCheck.h"
#include "GlobalTrace.h"
#include "GuiDialogOAC.h"
#include "MathTrig.h"
#include "MathRandom.h"


/*
 *  constructor: initialize all data members 
 */
TFDialogCraters::TFDialogCraters (HeightField *HF, HeightFieldDraw *HFD)
	        : TFPreviewDialog (HF, HFD, "Terraform Crater Dialog", 
					"Parameters", 6, 8),
		  d_vbList (TRUE, 0),
		  d_hbCX (TRUE, 5),
		  d_hbCY (TRUE, 5),
		  d_hbNumCraters (TRUE, 5),
		  d_hbRadius (TRUE, 5),
		  d_hbHeight (TRUE, 5),
		  d_hbCraterCoverage (TRUE, 5),
		  d_hbWrap (TRUE, 5),
		  d_hbGenSeed (TRUE, 5),
		  d_hbSeed (TRUE, 5),
		  d_optionsFrame ("Crater Options")
{
	GlobalTrace::trace (GlobalTrace::TRACE_FLOW, "+++ TFDialogCraters\n");
	d_seedInitStr = new char [40];
	d_lastPreviewSeed = -1;
	d_adjCX = new Gtk_Adjustment (0.5, 0, 1, 0.01);
	d_adjCY = new Gtk_Adjustment (0.5, 0, 1, 0.01);
	d_adjNumCraters = new Gtk_Adjustment (20, 1, 100, 0.01);
	d_adjRadius = new Gtk_Adjustment (0.33, 0, 3, 0.01);
	d_adjHeight = new Gtk_Adjustment (0.33, 0, 2, 0.01);
	d_adjCraterCoverage = new Gtk_Adjustment (0.15, 0.01, 0.5, 0.01);

	d_hsCX = new Gtk_HScale (*d_adjCX);
	d_hsCY = new Gtk_HScale (*d_adjCY);
	d_hsNumCraters = new Gtk_HScale (*d_adjNumCraters);
	d_hsRadius = new Gtk_HScale (*d_adjRadius);
	d_hsHeight = new Gtk_HScale (*d_adjHeight);
	d_hsCraterCoverage = new Gtk_HScale (*d_adjCraterCoverage);

	d_lblCX = new Gtk_Label ("Center X");
	d_lblCY = new Gtk_Label ("Center Y");
	d_lblNumCraters = new Gtk_Label ("Number of Craters");
	d_lblRadius = new Gtk_Label ("Crater Radius");
	d_lblHeight = new Gtk_Label ("Crater Height");
	d_lblCraterCoverage = new Gtk_Label ("CraterCoverage");
	d_lblWrap = new Gtk_Label ("Wrap");
	d_lblGenSeed = new Gtk_Label ("Generate new seed");
	d_lblSeed = new Gtk_Label ("Specity Seed");

	buildDialogWindow ();

	this->setHFobjs (HF, HFD);
	this->iterateEvents ();
	adjNumCratersCallback (); 	// calls updatePreviewCallback()

	connect_to_method (d_adjCX->value_changed, this, &TFDialogCraters::updatePreviewCallback);
	connect_to_method (d_adjCY->value_changed, this, &TFDialogCraters::updatePreviewCallback);
	connect_to_method (d_adjNumCraters->value_changed, this, &TFDialogCraters::adjNumCratersCallback);
	connect_to_method (d_adjRadius->value_changed, this, &TFDialogCraters::updatePreviewCallback);
	connect_to_method (d_adjHeight->value_changed, this, &TFDialogCraters::updatePreviewCallback);
	connect_to_method (d_adjCraterCoverage->value_changed, this, &TFDialogCraters::updatePreviewCallback);
	connect_to_method (d_btnWrap.clicked, this, &TFDialogCraters::updatePreviewCallback);
	connect_to_method (d_btnGenSeed.clicked, this, &TFDialogCraters::seedCheckboxCallback);
}


/*
 *  destructor: clean up 
 */
TFDialogCraters::~TFDialogCraters ()
{
	delete [] d_seedInitStr;

	delete d_adjCX;
	delete d_adjCY;
	delete d_adjNumCraters;
	delete d_adjRadius;
	delete d_adjHeight;
	delete d_adjCraterCoverage;

	delete d_hsCX;
	delete d_hsCY;
	delete d_hsNumCraters;
	delete d_hsRadius;
	delete d_hsHeight;
	delete d_hsCraterCoverage;

	delete d_lblCX;
	delete d_lblCY;
	delete d_lblNumCraters;
	delete d_lblRadius;
	delete d_lblHeight;
	delete d_lblCraterCoverage;
	delete d_lblWrap;
	delete d_lblGenSeed;
	delete d_lblSeed;
	GlobalTrace::trace (GlobalTrace::TRACE_FLOW, "--- TFDialogCraters\n");
}


/*
 *  insertOptions: fill the dialog's VBox
 */
void TFDialogCraters::insertOptions ()
{
	d_optionsFrame.set_shadow_type (GTK_SHADOW_ETCHED_IN);
	d_optionsFrame.set_border_width (5);

	d_hbCX.pack_start (*d_lblCX, TRUE, TRUE, GuiDialogOAC::s_HBOff);
	d_lblCX->show ();
	d_hsCX->set_update_policy (GTK_UPDATE_CONTINUOUS);
	d_hsCX->set_digits (2);
	d_hsCX->set_draw_value (TRUE);
	d_hbCX.pack_end (*d_hsCX, TRUE, TRUE, GuiDialogOAC::s_HBOff);
	d_hsCX->show ();
	d_vbList.pack_start (d_hbCX, TRUE, TRUE, GuiDialogOAC::s_VBOff);
	d_hbCX.show ();

	d_hbCY.pack_start (*d_lblCY, TRUE, TRUE, GuiDialogOAC::s_HBOff);
	d_lblCY->show ();
	d_hsCY->set_update_policy (GTK_UPDATE_CONTINUOUS);
	d_hsCY->set_digits (2);
	d_hsCY->set_draw_value (TRUE);
	d_hbCY.pack_end (*d_hsCY, TRUE, TRUE, GuiDialogOAC::s_HBOff);
	d_hsCY->show ();
	d_vbList.pack_start (d_hbCY, TRUE, TRUE, GuiDialogOAC::s_VBOff);
	d_hbCY.show ();

	d_vbList.pack_start (d_hSep);
	d_hSep.show ();

	d_hbNumCraters.pack_start (*d_lblNumCraters, TRUE, TRUE, GuiDialogOAC::s_HBOff);
	d_lblNumCraters->show ();
	d_hsNumCraters->set_update_policy (GTK_UPDATE_CONTINUOUS);
	d_hsNumCraters->set_digits (0);
	d_hsNumCraters->set_draw_value (TRUE);
	d_hbNumCraters.pack_end (*d_hsNumCraters, TRUE, TRUE, GuiDialogOAC::s_HBOff);
	d_hsNumCraters->show ();
	d_vbList.pack_start (d_hbNumCraters, TRUE, TRUE, GuiDialogOAC::s_VBOff);
	d_hbNumCraters.show ();

	d_hbRadius.pack_start (*d_lblRadius, TRUE, TRUE, GuiDialogOAC::s_HBOff);
	d_lblRadius->show ();
	d_hsRadius->set_update_policy (GTK_UPDATE_CONTINUOUS);
	d_hsRadius->set_digits (2);
	d_hsRadius->set_draw_value (TRUE);
	d_hbRadius.pack_end (*d_hsRadius, TRUE, TRUE, GuiDialogOAC::s_HBOff);
	d_hsRadius->show ();
	d_vbList.pack_start (d_hbRadius, TRUE, TRUE, GuiDialogOAC::s_VBOff);
	d_hbRadius.show ();

	d_hbHeight.pack_start (*d_lblHeight, TRUE, TRUE, GuiDialogOAC::s_HBOff);
	d_lblHeight->show ();
	d_hsHeight->set_update_policy (GTK_UPDATE_CONTINUOUS);
	d_hsHeight->set_digits (2);
	d_hsHeight->set_draw_value (TRUE);
	d_hbHeight.pack_end (*d_hsHeight, TRUE, TRUE, GuiDialogOAC::s_HBOff);
	d_hsHeight->show ();
	d_vbList.pack_start (d_hbHeight, TRUE, TRUE, GuiDialogOAC::s_VBOff);
	d_hbHeight.show ();

	d_hbCraterCoverage.pack_start (*d_lblCraterCoverage, TRUE, TRUE, GuiDialogOAC::s_HBOff);
	d_lblCraterCoverage->show ();
	d_hsCraterCoverage->set_update_policy (GTK_UPDATE_CONTINUOUS);
	d_hsCraterCoverage->set_digits (2);
	d_hsCraterCoverage->set_draw_value (TRUE);
	d_hbCraterCoverage.pack_end (*d_hsCraterCoverage, TRUE, TRUE, GuiDialogOAC::s_HBOff);
	d_hsCraterCoverage->show ();
	d_vbList.pack_start (d_hbCraterCoverage, TRUE, TRUE, GuiDialogOAC::s_VBOff);
	d_hbCraterCoverage.show ();

	d_hbWrap.pack_start (*d_lblWrap, TRUE, TRUE, GuiDialogOAC::s_HBOff);
	d_lblWrap->show ();
	d_hbWrap.pack_end (d_btnWrap, TRUE, TRUE, GuiDialogOAC::s_HBOff);
	d_btnWrap.show ();
	d_vbList.pack_start (d_hbWrap, TRUE, TRUE, GuiDialogOAC::s_VBOff);
	d_hbWrap.show ();

	d_vbList.pack_start (d_hSepSeed);
	d_hSepSeed.show ();

	// random seed parameters
	d_hbGenSeed.pack_start (*d_lblGenSeed, TRUE, TRUE, GuiDialogOAC::s_HBOff);
	d_lblGenSeed->show ();
	d_hbGenSeed.pack_end (d_btnGenSeed, TRUE, TRUE, GuiDialogOAC::s_HBOff);
	d_btnGenSeed.set_active (TRUE);
	d_btnGenSeed.show ();
	d_vbList.pack_start (d_hbGenSeed, TRUE, TRUE, GuiDialogOAC::s_VBOff);
	d_hbGenSeed.show ();

	d_lblSeed->set_justify (GTK_JUSTIFY_RIGHT);
	d_hbSeed.pack_start (*d_lblSeed, TRUE, TRUE, GuiDialogOAC::s_HBOff);
	d_lblSeed->set_sensitive (FALSE);
	d_lblSeed->show ();

	MathRandom	*mRand = new MathRandom ();
	sprintf (d_seedInitStr, "%d", (int)mRand->rnd());
	delete mRand;
	d_enSeed.set_text (d_seedInitStr);
	d_enSeed.set_sensitive (FALSE);
	//d_hbSeed.pack_start (d_enSeed, TRUE, TRUE, GuiDialogOAC::s_HBOff);
	d_hbSeed.pack_start (d_enSeed, TRUE, TRUE, 0);
	d_enSeed.show ();

	d_vbList.pack_start (d_hbSeed, TRUE, TRUE, GuiDialogOAC::s_VBOff);
	d_hbSeed.show ();


	d_optionsFrame.add (d_vbList);
	this->d_table->attach (d_optionsFrame, 3, 6, 1, 7);
        d_vbList.show ();
	d_optionsFrame.show ();
        this->d_baseFrame->show ();
        this->get_vbox()->show ();
}



void TFDialogCraters::buildDialogWindow ()
{
	this->set_usize (450, 450);
	this->set_title (this->d_winTitle);

	this->get_vbox()->set_border_width (2);

	this->setupVBox ();
	this->insertPreview ();
	insertOptions ();
	this->fillActionArea ();

	connect_to_method (d_drawArea->button_press_event, this, 
			&TFDialogCraters::daButtonPressedCallback);
	
	this->show ();
}


void TFDialogCraters::seedCheckboxCallback ()
{
	bool	status=d_enSeed.is_sensitive ();

	d_lblSeed->set_sensitive (!status);
	d_enSeed.set_sensitive (!status);
}


/*
 *  buttonCallbackApply: roughen the Height Field
 */
void TFDialogCraters::buttonCallbackApply ()
{
	float			cx, cy, nc, radius, height, craterCoverage;
	int			seed;
	bool			wrap;

	cx = d_adjCX->get_value ();
	cy = d_adjCY->get_value ();
        nc = d_adjNumCraters->get_value ();
        radius = d_adjRadius->get_value (); 
        height = d_adjHeight->get_value (); 
        craterCoverage = d_adjCraterCoverage->get_value (); 
	wrap = d_btnWrap.get_active ();

	if (d_btnGenSeed.get_active())
		{
		MathRandom      *mRand = new MathRandom ();

		seed = (int) (mRand->rnd()*G_MAXINT);
		sprintf (d_seedInitStr, "%d", seed);
		d_enSeed.set_text (d_seedInitStr);
		delete mRand;
		}
	else
		{
		char *s = const_cast<char*>(d_enSeed.get_text().c_str());
		seed = atoi (s);
		}

	SanityCheck::bailout ((!d_HFO), "d_HFO==NULL", "TFDialogCraters::buttonCallbackApply");
	if (d_adjNumCraters->get_value() == 1.0)
		d_HFO->distributeCraters (nc, wrap, height, radius, craterCoverage, seed, cx, cy);
	else
		d_HFO->distributeCraters (nc, wrap, height, radius, craterCoverage, seed);
	d_HFD->draw ();

	// only do this if window stays open 
	if (d_applyHit)
		{
		this->setHFobjs (d_HF, d_HFD);
		updatePreviewCallback ();
		}

	d_lastPreviewSeed = -1;
	this->d_applyHit = TRUE;
}


/*
 *  updatePreviewCallback: update the preview after a slider has been updated.
 */
void TFDialogCraters::updatePreviewCallback ()
{
	if (!d_cbUsePreview.get_active())
		return;

	float		cx, cy, nc, radius, height, craterCoverage;
	int		seed;
	bool		wrap;

	cx = d_adjCX->get_value ();
	cy = d_adjCY->get_value ();
        nc = d_adjNumCraters->get_value ();
        radius = d_adjRadius->get_value (); 
        height = d_adjHeight->get_value (); 
        craterCoverage = d_adjCraterCoverage->get_value (); 
	wrap = d_btnWrap.get_active ();

	if (wrap)
		nc = 1-(nc-1);

	if (d_btnGenSeed.get_active())
		{
		if (d_lastPreviewSeed == -1)
			{
			MathRandom      *mRand = new MathRandom ();

			seed = d_lastPreviewSeed = (int) (mRand->rnd()*G_MAXINT);
			sprintf (d_seedInitStr, "%d", seed);
			d_enSeed.set_text (d_seedInitStr);
			delete mRand;
			}
		else
			seed = d_lastPreviewSeed;
		}
	else
		{
		char *s = const_cast<char*>(d_enSeed.get_text().c_str());
		seed = d_lastPreviewSeed = atoi (s);
		}

	SanityCheck::bailout ((!d_HFPreview), "d_HFPreview==NULL", "TFDialogCraters::updatePreviewCallback");
	SanityCheck::bailout ((!d_HFOPreview), "d_HFOPreview==NULL", "TFDialogCraters::updatePreviewCallback");
	SanityCheck::bailout ((!d_HFDPreview), "d_HFOPreview==NULL", "TFDialogCraters::updatePreviewCallback");
	this->previewUpToDate ();
	d_HFPreview->restore ();
	if (d_adjNumCraters->get_value() == 1.0)
		d_HFOPreview->distributeCraters (nc, wrap, height, radius, craterCoverage, seed, cx, cy);
	else
		d_HFOPreview->distributeCraters (nc, wrap, height, radius, craterCoverage, seed);
	d_HFDPreview->setColormap (d_HFD->getColormap());
	d_HFDPreview->draw ();
	d_cx = (int)(cx*(float)d_drawArea->width()); 
	d_cy = (int)(cy*(float)d_drawArea->height()); 
	if (d_adjNumCraters->get_value() == 1.0)
		this->drawCrosshairs (d_cx, d_cy);
}


/*
 *  daButtonPressedCallback: update the preview d_drawArea and adjustment 
 * 	the appropriate widgets (done when the user clicks on the preview)
 */
int TFDialogCraters::daButtonPressedCallback (GdkEventButton *e)
{
	if (!d_cbUsePreview.get_active())
		return 1;

	if (d_adjNumCraters->get_value() != 1.0)
		return 1;

	int	w = this->d_drawArea->width(),
		h = this->d_drawArea->height();

	if (e->button == 1)
		{
		d_adjCX->set_value (e->x/w);
		d_adjCY->set_value (e->y/h);
		}
	else
	if (e->button == 3)
		{
		float	d = MathTrig::distance (d_adjCX->get_value()*w, 
				d_adjCY->get_value()*h, e->x, e->y);

		d/=((d_drawArea->width()+d_drawArea->height())/2);
		if (d > 1)
			d=1;
		d_adjRadius->set_value (d); 
		}
	updatePreviewCallback ();
	return 0;
}



/*
 *  adjNumCratersCallback: (de)activate when NumCrates slider==1 or when 
 * 	it was 1 and is now a different value
 */
void TFDialogCraters::adjNumCratersCallback ()
{
	if (d_adjNumCraters->get_value() == 1.0)
		{
		if (!d_lblCX->is_sensitive())
			{
			d_lblCX->set_sensitive (TRUE);
			d_lblCY->set_sensitive (TRUE);
			d_hsCX->set_sensitive (TRUE);
			d_hsCY->set_sensitive (TRUE);
			}
		}
	else
	if (d_lblCX->is_sensitive())
		{
		d_lblCX->set_sensitive (FALSE);
		d_lblCY->set_sensitive (FALSE);
		d_hsCX->set_sensitive (FALSE);
		d_hsCY->set_sensitive (FALSE);
		}
	updatePreviewCallback ();
}



/*
 *  (overloaded virtual) checkboxCallbackPreview: process the 'use preview' 
 * 	checkbox. This is a special case to only draw crosshairs when 
 * 	number of craters == 1
 */
void TFDialogCraters::checkboxCallbackPreview ()
{
	if (!d_cbUsePreview.get_active())
		{
		d_HFPreview->restore ();
		d_HFDPreview->draw ();
		}
	else
		{
		SanityCheck::bailout ((!d_HFDPreview), "HFDPreview==NULL", 
				"GuiDialogPreview::checkboxCallbackPreview");
		d_drawArea->set_sensitive (TRUE);	
		updatePreviewCallback ();
		d_HFDPreview->draw ();
		if ( (d_cx!=-1) && (d_cy!=-1) && d_adjNumCraters->get_value()==1)
			drawCrosshairs (d_cx, d_cy);
		}
}


